// Device-specific include file.
#include <p33fxxxx.h>

// VDC definitions.
#include "vdc.h"
#include "vdc_graphics.h"
#include "vdc_text.h"

// Stores a 320x240 monochrome bitmap.
unsigned int __attribute__((far,noload)) vdc_frame_buffer[4800];

// Stores a single scanline temporarily (used for DMA transfers).
unsigned int __attribute__((space(dma),noload)) vdc_scanline_buffer[21];

// Points at the current part of the display that's being sent to the display device.
volatile unsigned int* vdc_frame_scan_ptr;

// Stores the current scanline number.
volatile int vdc_active_line_counter;

// Stores the current frame rate of the display refresh.
int vdc_fps;

// Pointer to the current LCD/TV retrace function.
volatile unsigned int vdc_retrace_function_ptr;

// (TV specific) Counter for the number of times to repeat the current TV retrace function.
volatile unsigned int vdc_tv_retrace_function_repeat_count;

// (TV specific) Offsets the display in the X axis.
int vdc_tv_line_offset_x = 82;

// (TV specific) Offsets the display in the Y axis.
int vdc_tv_line_offset_y = 35;

// (TV specific) Initialises the TV retracing routine.
extern void vdc_tv_retrace_init(void);

// (LCD specific) Initialises the LCD retracing routine.
extern void vdc_lcd_retrace_init(void);

/// (VGA specific) Initialises the VGA retracing routine.
extern void vdc_vga_retrace_init(void);

// Stores the current output mode.
enum vdc_output_mode_t vdc_output_mode = -1;

// (TV specific) Cleared if the TV is not detected during a retrace.
volatile int vdc_tv_connected;

// (VGA specific) Cleared if the VGA monitor is not detected during a retrace.
volatile int vdc_vga_connected;

// Pin and port definitions
#define TV_LATCH LATB
#define TV_TRIS TRISB
#define TV_PORT PORTB
#define TV_SYNC 6
#define TV_PICTURE 5
#define TV_PIXEL_CLOCK 4
#define TV_DETECT 7

#define VGA_LATCH LATB
#define VGA_TRIS TRISB
#define VGA_PORT PORTB
#define VGA_HSYNC 1
#define VGA_VSYNC 2
#define VGA_PIXEL_CLOCK 4
#define VGA_PICTURE 5
#define VGA_DETECT 10

// Resets both graphics and text systems. This does not (re)initialise output.
void vdc_reset(void) {
	vdc_graphics_reset();
	vdc_text_reset();
}

// De-initialise the TV output code.
void vdc_tv_deinit(void) {

	T1CONbits.TON = 0; // Switch timer 1 off.
	T2CONbits.TON = 0; // Switch timer 2 off.

	OC1CONbits.OCM = 0b000; // Switch output capture unit 1 off.

	// Stop generating timer 1 interrupts.
	IEC0bits.T1IE = 0; // Disable the interrupt.
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	
	// Switch off the two DMA channels.
	DMA0CONbits.CHEN = 0;
	DMA1CONbits.CHEN = 0;

	_RP4R = 0b00000; // Return pin RP4 to its default use.
	_RP5R = 0b00000; // Return pin RP5 to its default use.
	RPINR20bits.SCK1R = 0b1111; // Pull SCK1 high.

	// Switch off SPI1.
	SPI1STATbits.SPIEN = 0;

	vdc_tv_connected = 0;
}

// De-initialise the LCD output code.
void vdc_lcd_deinit(void) {

	// Switch the LCD off.
	LATB = 0;
	
	T1CONbits.TON = 0; // Switch timer 1 off.

	// Stop generating timer 1 interrupts.
	IEC0bits.T1IE = 0; // Disable the interrupt.
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	
	// Disconnect the LCD data pins.
	TRISB |= 0b1;
	TRISA |= 0b10011;	
}

// De-initialise the VGA output code.
void vdc_vga_deinit(void) {
	vdc_tv_deinit();
	VGA_TRIS |= (1 << VGA_HSYNC) | (1 << VGA_VSYNC);
	vdc_vga_connected = 0;
	
}

// Initialise the TV output code.
void vdc_tv_init(void) {
	
	// Ensure a suitable retrace function is loaded.
	vdc_tv_retrace_init();

	// Reset the scan pointer.
	vdc_frame_scan_ptr = vdc_frame_buffer;
	vdc_active_line_counter = 0;

	TV_TRIS &= ~(1 << TV_PIXEL_CLOCK);

	// Configure timer 1 to trigger interrupts to constantly retrace the TV.
	T1CONbits.TCKPS = 0b00; // No prescaler.
	PR1 = (FCY / 1000000UL * 64) - 1; // Trigger the interrupt when the counter reaches this value.
	TMR1 = 0; // Reset the timer.
	T1CONbits.TON = 1; // Switch on the timer.

	// Configure timer 2 to generate the clock
	TMR2 = 0;
	PR2 = 4;
	T2CONbits.TON = 1;

	OC1CONbits.OCTSEL = 0;
	OC1CONbits.OCM = 0b101;
	OC1R = 0;
	OC1RS = 2;
	_RP4R = 0b10010; // Output OC1 on pin RP4.
	
	SPI1CON1bits.MSTEN = 0; // Slave transmitter mode.

	SPI1CON1bits.MODE16 = 1; // 16-bit mode.

	SPI1CON1bits.DISSCK = 1; // Disable the SCK pin.
		
	_RP5R = 0b00111; // Output SDO1 on pin RP5.
	RPINR20bits.SCK1R = TV_PIXEL_CLOCK; // Input SCK1 on pin RPx.

	DMACS0 = 0;

	DMA0REQbits.IRQSEL = 0b0001010;
	DMA0PAD = (unsigned int)&SPI1BUF;
	DMA0CONbits.DIR = 1;
	DMA0CONbits.MODE = 0b01;
	DMA0STA = __builtin_dmaoffset(vdc_scanline_buffer);
	DMA0CNT = 20;

	DMA1REQbits.IRQSEL = 0b0001010;
	DMA1PAD = (unsigned int)&SPI1BUF;
	DMA1CONbits.DIR = 0;
	DMA1CONbits.MODE = 0b01;
	DMA1STA = __builtin_dmaoffset(vdc_scanline_buffer);
	DMA1CNT = 20;
	
	// Enable the timer 1 interrupt.
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	IEC0bits.T1IE = 1; // Enable the interrupt.

	vdc_fps = 50;
	
	vdc_output_mode = VDC_OUTPUT_TV;

	// Assume a TV is connected.
	vdc_tv_connected = !0;
}

// Initialise the LCD output code.
void vdc_lcd_init(void) {

	// Ensure a suitable retrace function is loaded.
	vdc_lcd_retrace_init();

	// Reset the scan pointer.
	vdc_frame_scan_ptr = vdc_frame_buffer;
	vdc_active_line_counter = 0;

	// Force the LCD control pins low and make the LCD data pins outputs.
	LATB = 0;
	TRISB &= ~0b1;
	TRISA &= ~0b10011;

	// Configure timer 1 to trigger interrupts to constantly retrace the LCD.
	T1CONbits.TCKPS = 0b00; // No prescaler.
	PR1 = (((unsigned long)FCY / (240)) / 55) - 1; // Trigger the interrupt when the counter reaches this value.
	
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	IEC0bits.T1IE = 1; // Enable the interrupt.
	
	TMR1 = 0; // Reset the timer.
	T1CONbits.TON = 1; // Switch on the timer.

	// Switch the LCD on.
	LATB |= (1 << 11);

	vdc_fps = 55;

	vdc_output_mode = VDC_OUTPUT_LCD;

}

// Initialise the VGA output code.
void vdc_vga_init(void) {
	
	// Ensure a suitable retrace function is loaded.
	vdc_vga_retrace_init();

	// Reset the scan pointer.
	vdc_frame_scan_ptr = vdc_frame_buffer;
	vdc_active_line_counter = 0;

	VGA_TRIS &= ~((1 << VGA_PIXEL_CLOCK) | (1 << VGA_HSYNC) | (1 << VGA_VSYNC));

	// Configure timer 1 to trigger interrupts to constantly retrace the TV.
	T1CONbits.TCKPS = 0b00; // No prescaler.
	PR1 = (FCY / 1000000UL * 64) - 1; // Trigger the interrupt when the counter reaches this value.
	TMR1 = 0; // Reset the timer.
	T1CONbits.TON = 1; // Switch on the timer.

	// Configure timer 2 to generate the clock
	TMR2 = 0;
	PR2 = 2;
	T2CONbits.TON = 1;

	OC1CONbits.OCTSEL = 0;
	OC1CONbits.OCM = 0b101;
	OC1R = 0;
	OC1RS = 2;
	_RP4R = 0b10010; // Output OC1 on pin RP4.
	
	SPI1CON1bits.MSTEN = 0; // Slave transmitter mode.

	SPI1CON1bits.MODE16 = 1; // 16-bit mode.

	SPI1CON1bits.DISSCK = 1; // Disable the SCK pin.
		
	_RP5R = 0b00111; // Output SDO1 on pin RP5.
	RPINR20bits.SCK1R = VGA_PIXEL_CLOCK; // Input SCK1 on pin RPx.

	DMACS0 = 0;

	DMA0REQbits.IRQSEL = 0b0001010;
	DMA0PAD = (unsigned int)&SPI1BUF;
	DMA0CONbits.DIR = 1;
	DMA0CONbits.MODE = 0b01;
	DMA0STA = __builtin_dmaoffset(vdc_scanline_buffer);
	DMA0CNT = 20;

	DMA1REQbits.IRQSEL = 0b0001010;
	DMA1PAD = (unsigned int)&SPI1BUF;
	DMA1CONbits.DIR = 0;
	DMA1CONbits.MODE = 0b01;
	DMA1STA = __builtin_dmaoffset(vdc_scanline_buffer);
	DMA1CNT = 20;
	
	// Enable the timer 1 interrupt.
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	IEC0bits.T1IE = 1; // Enable the interrupt.

	vdc_fps = 60;
	
	vdc_output_mode = VDC_OUTPUT_VGA;

	// Assume a VGA monitor is connected.
	vdc_vga_connected = !0;

}

// Initialises the VDC output.
void vdc_init_output(enum vdc_output_mode_t output_mode) {
	
	// Ensure that LCD control pins and TV pins are driven low.
	LATB &= 0b0000011110011111;
	TRISB &= 0b0000011110011111;
	
	// Set interrupt priorities.
	IPC0bits.T1IP = 7; // Timer 1 has highest priority.
	IPC2bits.SPI1EIP = 6; // SPI1 has second-highest priority.

	// De-initialise the old mode.
	switch (vdc_output_mode) {
		case VDC_OUTPUT_TV:
			vdc_tv_deinit();
			break;
		case VDC_OUTPUT_LCD:
			vdc_lcd_deinit();
			break;
		case VDC_OUTPUT_VGA:
			vdc_vga_deinit();
			break;
	}
	
	// Initialise the new mode.
	switch (output_mode) {
		case VDC_OUTPUT_TV:
			vdc_tv_init();
			break;
		case VDC_OUTPUT_LCD:
			vdc_lcd_init();
			break;
		case VDC_OUTPUT_VGA:
			vdc_vga_init();
			break;
	}

}

// Checks whether a TV is connected or not. May only be updated once a frame.
int vdc_tv_is_connected(void) {
	int result = 0;
	if (vdc_output_mode == VDC_OUTPUT_LCD) {
		// We can directly test the pin as we're not doing anything with the external display at the moment.
		TV_LATCH |= (1 << TV_SYNC) | (1 << TV_PICTURE);
		Nop();
		Nop();
		Nop();
		result = !(TV_PORT & (1 << TV_DETECT));
		// Drive the TV_SYNC and TV_PICTURE pins low.
		TV_LATCH &= ~((1 << TV_SYNC) | (1 << TV_PICTURE));
	} else {
		// We can't directly test the pin so use the value updated by the retrace functions.
		result = vdc_tv_connected;
	}
	return result;
}

// Checks whether a VGA monitor is connected or not. May only be updated once a frame.
int vdc_vga_is_connected(void) {
	int result = 0;
	if (vdc_output_mode == VDC_OUTPUT_LCD) {
		// We can directly test the pin as we're not doing anything with the external display at the moment.
		VGA_LATCH |= (1 << VGA_PICTURE);
		Nop();
		Nop();
		Nop();
		result = !(VGA_PORT & (1 << VGA_DETECT));
		// Drive the VGA_PICTURE pin low.
		VGA_LATCH &= ~(1 << VGA_PICTURE);
	} else {
		// We can't directly test the pin so use the value updated by the retrace functions.
		result = vdc_vga_connected;
	}
	return result;
}

// Automatically (re)initises the VDC to output to the TV (if connected) or the LCD (if not).
void vdc_auto_init_output(void) {
	if (vdc_tv_is_connected()) {
		if (vdc_output_mode != VDC_OUTPUT_TV) vdc_init_output(VDC_OUTPUT_TV);
	} else if (vdc_vga_is_connected()) {
		if (vdc_output_mode != VDC_OUTPUT_VGA) vdc_init_output(VDC_OUTPUT_VGA);
	} else {
		if (vdc_output_mode != VDC_OUTPUT_LCD) vdc_init_output(VDC_OUTPUT_LCD);
	}
}

// Waits for the display to finish retracing and enter the vertical blank region.
void vdc_wait_vblank(void) {
	vdc_auto_init_output();
	switch (vdc_output_mode) {
		case VDC_OUTPUT_TV:
			while (vdc_active_line_counter != 240);
			break;
		case VDC_OUTPUT_LCD:
			while (vdc_active_line_counter);
			break;
		case VDC_OUTPUT_VGA:
			while (vdc_active_line_counter != 480);
			break;
	}
}

// Delay for a particular number of frames.
void vdc_delay_frames(int frames) {
	while (frames--) {
		vdc_auto_init_output();
		switch (vdc_output_mode) {
			case VDC_OUTPUT_TV:
				while (vdc_active_line_counter != 240);
				while (vdc_active_line_counter == 240);
				break;
			case VDC_OUTPUT_LCD:
				while (vdc_active_line_counter);
				while (!vdc_active_line_counter);
				break;
			case VDC_OUTPUT_VGA:
				while (vdc_active_line_counter != 480);
				while (vdc_active_line_counter == 480);
				break;
		}
	}
}
