// Device-specific include file.
#include <p33fxxxx.h>

// VDC definitions.
#include "vdc.h"
#include "vdc_text.h"
#include "vdc_text_font.h"

// Stores a 80x30 text character buffer.
char __attribute__((far,noload)) vdc_text_buffer[80 * 30];

// Defines the current text viewport.
struct vdc_rectangle_t vdc_text_viewport = { 0, 0, 79, 29 };

// Stores the current text cursor position.
struct vdc_point_2d_t vdc_text_cursor_position = { 0, 0 };

// Stores the text foreground colour.
unsigned int vdc_text_foreground_colour = 127;

// Stores the text background colour.
unsigned int vdc_text_background_colour = 0;

// Defines the direction of a text scroll.
enum vdc_text_scroll_direction_t {
	VDC_TEXT_SCROLL_UP,
	VDC_TEXT_SCROLL_DOWN,
};

// Resets the font size to the small font, resets the text viewport, sets the background colour to black, the foreground colour to white and clears the screen.
void vdc_text_reset(void) {
	vdc_text_foreground_colour = 127;
	vdc_text_background_colour = 0;
	vdc_text_set_font_size(VDC_TEXT_FONT_SIZE_SMALL);
}

// Clears the text viewport.
void vdc_text_clear(void) {
	for (vdc_text_cursor_position.y = vdc_text_viewport.bottom; vdc_text_cursor_position.y >= vdc_text_viewport.top; --vdc_text_cursor_position.y) {
		for (vdc_text_cursor_position.x = vdc_text_viewport.right; vdc_text_cursor_position.x >= vdc_text_viewport.left; --vdc_text_cursor_position.x) {	
			vdc_text_put_map(' ');
		}
	}
	++vdc_text_cursor_position.x;
	++vdc_text_cursor_position.y;
}

// Resets the text viewport to fill the screen.
void vdc_text_reset_viewport(void) {
	vdc_text_viewport.left = 0;
	vdc_text_viewport.top = 0;
	vdc_text_viewport.bottom = 29;
	vdc_text_viewport.right = (320 / vdc_text_font_size) - 1;
}

// Puts a character on the display at the current cursor position.
void vdc_text_put_map(char c) {
	
	// Write the character to the text buffer.
	vdc_text_buffer[vdc_text_cursor_position.x + vdc_text_cursor_position.y * 80] = c;
	
	// Write the character to the display.
	
	unsigned int character_row;
	unsigned int character_row_data;
	unsigned int character_row_data_masked;

	unsigned int* font_data_source = vdc_font + c * 4;
	unsigned int* frame_buffer_destination = vdc_frame_buffer + (vdc_text_cursor_position.y * 20 * 8) + (vdc_text_cursor_position.x / (16 / vdc_text_font_size));
	
	for (character_row = 0; character_row < 8; ++character_row) {
		// Clear the background of the cell.
		switch (vdc_text_font_size) {
			case VDC_TEXT_FONT_SIZE_LARGE:
				if ((vdc_text_cursor_position.x & 1) == 0) {
					if (vdc_text_background_colour < 64) {
						*frame_buffer_destination &= 0x00FF;
					} else {
						*frame_buffer_destination |= 0xFF00;
					}
				} else {
					if (vdc_text_background_colour < 64) {
						*frame_buffer_destination &= 0xFF00;
					} else {
						*frame_buffer_destination |= 0x00FF;
					}
				}
				break;
			case VDC_TEXT_FONT_SIZE_SMALL:
				if (vdc_text_background_colour < 64) {
					*frame_buffer_destination &= ~(0xF000 >> ((vdc_text_cursor_position.x & 3) * 4));
				} else {
					*frame_buffer_destination |= (0xF000 >> ((vdc_text_cursor_position.x & 3) * 4));
				}
				break;
		}


		// Get the character row data from the font.
		switch (vdc_text_font_size) {
			case VDC_TEXT_FONT_SIZE_LARGE:
				if ((character_row & 1) == 0) {
					// Fetch the raw character font data from RAM.
					character_row_data = *font_data_source++;
					// If the foreground and background colours match, use a blank character.
					if ((vdc_text_foreground_colour < 64) == (vdc_text_background_colour < 64)) character_row_data = 0x0000;
					// Do we need to shunt the data along or merely mask it for the current cell?
					if ((vdc_text_cursor_position.x & 1) == 0) {
						character_row_data_masked = character_row_data & 0xFF00;
					} else {
						character_row_data_masked = character_row_data >> 8;
					}
				} else {
					// Do we need to shunt the data along or merely mask it for the current cell?
					if ((vdc_text_cursor_position.x & 1) == 0) {
						character_row_data_masked = character_row_data << 8;
					} else {
						character_row_data_masked = character_row_data & 0x00FF;
					}
				}
				break;
			case VDC_TEXT_FONT_SIZE_SMALL:
				if ((character_row & 1) == 0) {
					// Fetch the raw character font data from RAM.
					character_row_data = *font_data_source++;
					// If the foreground and background colours match, use a blank character.
					if ((vdc_text_foreground_colour < 64) == (vdc_text_background_colour < 64)) character_row_data = 0x0000;
				}
				character_row_data_masked = ((character_row_data << (character_row & 1) * 8) & 0xF000) >> ((vdc_text_cursor_position.x & 3) * 4);
				break;
		}
		// Write the character data to the frame buffer.
		*frame_buffer_destination ^= character_row_data_masked;
		// Advance to the next row of the frame buffer.
		frame_buffer_destination += 20;
	}
	
}

// Scrolls the viewport up or down one row.
void vdc_text_scroll_up_down(enum vdc_text_scroll_direction_t direction) {
	
	int x, y;

	int min_x, max_x;

	unsigned int* source;
	unsigned int* destination;

	unsigned int* source_c;
	unsigned int* destination_c;

	unsigned int left_mask, right_mask, mask;

	int stride;
	unsigned int rows_to_shift;

	// Are we scrolling up or down?
	switch (direction) {
		case VDC_TEXT_SCROLL_UP:
			stride = +20;
			break;
		case VDC_TEXT_SCROLL_DOWN:
			stride = -20;
			break;
	}

	// Calculate min and max values for data aligned to 16 bit boundaries.
	switch (vdc_text_font_size) {
		case VDC_TEXT_FONT_SIZE_LARGE:
			min_x = vdc_text_viewport.left / 2;
			max_x = vdc_text_viewport.right / 2;
			left_mask = (unsigned)0xFFFF >> ((vdc_text_viewport.left & 1) * 8);
			right_mask = (signed)0xFF00 >> ((vdc_text_viewport.right & 1) * 8);
			break;
		case VDC_TEXT_FONT_SIZE_SMALL:
			min_x = vdc_text_viewport.left / 4;
			max_x = vdc_text_viewport.right / 4;
			left_mask = (unsigned)0xFFFF >> ((vdc_text_viewport.left & 3) * 4);
			right_mask = (signed)0xF000 >> ((vdc_text_viewport.right & 3) * 4);
			break;
	}
	
	// Source and destination pointers to shunt the bitmap data up.
	switch (direction) {
		case VDC_TEXT_SCROLL_UP:
			destination = vdc_frame_buffer + vdc_text_viewport.top * 8 * 20 + min_x;
			source = destination + 8 * 20;
			break;
		case VDC_TEXT_SCROLL_DOWN:
			destination = vdc_frame_buffer + (vdc_text_viewport.bottom * 8 + 7) * 20 + min_x;
			source = destination - 8 * 20;
			break;
	}

	// How many rows do we need to shift?
	rows_to_shift = (vdc_text_viewport.bottom - vdc_text_viewport.top) * 8;

	if (min_x == max_x) {
		// Special case where source and destination lie in the same integer.
		
		mask = left_mask & right_mask;

		// Shunt up each row.
		for (y = 0; y < rows_to_shift; ++y, source += stride, destination += stride) {
			*destination &= ~mask;
			*destination |= *source & mask;
		}
		// Clear the top or bottom row.
		for (y = 0; y < 8; ++y, destination += stride) {
			if (vdc_text_background_colour < 64) {
				*destination &= ~mask;
			} else {
				*destination |= mask;
			}
		}
	} else {
		
		// Shunt the frame buffer up or down.
		for (y = 0; y < rows_to_shift; ++y, source += stride, destination += stride) {
			
			// Per-column copies of the source and destination pointers.
			source_c = source;
			destination_c = destination;
			
			// Left-over left column.
			*(destination_c) &= ~left_mask;
			*(destination_c++) |= (*(source_c++) & left_mask);
			
			// Central part that can be copied word-by-word.
			for (x = min_x + 1; x < max_x; ++x) {
				*destination_c++ = *source_c++;
			}
	
			// Left-over right column.			
			*(destination_c) &= ~right_mask;
			*(destination_c) |= (*(source_c) & right_mask);
	
		}
	
		// Clear the bottom row.
		
		for (y = 0; y < 8; ++y, destination += stride) {
			
			// Per-column copies of the destination pointer.
			destination_c = destination;

			// Left-over left column.			
			if (vdc_text_background_colour < 64) {
				*(destination_c++) &= ~left_mask;
			} else {
				*(destination_c++) |= left_mask;
			}
			
			// Central part that can be erased word-by-word.
			for (x = min_x + 1; x < max_x; ++x) {
				if (vdc_text_background_colour < 64) {
					*destination_c++ = 0x0000;
				} else {
					*destination_c++ = 0xFFFF;
				}
			}
	
			// Left-over right column.
			if (vdc_text_background_colour < 64) {
				*(destination_c) &= ~right_mask;
			} else {
				*(destination_c) |= right_mask;
			}
		}
	}


	// Scroll the text buffer up or down.
	char* text_buffer_destination;
	char* text_buffer_source;
	char* text_buffer_destination_c;
	char* text_buffer_source_c;

	// Are we scrolling up or down?
	switch (direction) {
		case VDC_TEXT_SCROLL_UP:
			stride = +80;
			text_buffer_destination = vdc_text_buffer + vdc_text_viewport.top * 80 + vdc_text_viewport.left;
			text_buffer_source = text_buffer_destination + 80;
			break;
		case VDC_TEXT_SCROLL_DOWN:
			stride = -80;
			text_buffer_destination = vdc_text_buffer + vdc_text_viewport.bottom * 80 + vdc_text_viewport.left;
			text_buffer_source = text_buffer_destination - 80;
			break;
	}

	rows_to_shift = vdc_text_viewport.bottom - vdc_text_viewport.top;

	for (y = 0; y < rows_to_shift; ++y, text_buffer_destination += stride, text_buffer_source += stride) {
		text_buffer_destination_c = text_buffer_destination;
		text_buffer_source_c = text_buffer_source;
		for (x = vdc_text_viewport.left; x <= vdc_text_viewport.right; ++x) {
			*text_buffer_destination_c++ = *text_buffer_source_c++;
		}
	}

	// Clear the bottom row of the text buffer.
	text_buffer_destination_c = text_buffer_source;
	for (x = vdc_text_viewport.left; x <= vdc_text_viewport.right; ++x) {
		*text_buffer_destination_c++ = ' ';
	}
	
}

// Scrolls the viewport up one row.
void vdc_text_scroll_up(void) {
	vdc_text_scroll_up_down(VDC_TEXT_SCROLL_UP);
}

// Scrolls the viewport down one row.
void vdc_text_scroll_down(void) {
	vdc_text_scroll_up_down(VDC_TEXT_SCROLL_DOWN);
}

// Moves the text cursor down a row. If it moves past the bottom edge of the viewport the cursor returns to the bottom edge of the viewport and the viewport scrolls up one row.
void vdc_text_move_cursor_down(void) {
	if (++vdc_text_cursor_position.y > vdc_text_viewport.bottom) {
		vdc_text_cursor_position.y = vdc_text_viewport.bottom;
		vdc_text_scroll_up();
	}
}

// Moves the text cursor right a column. If it moves past the right edge of the viewport the cursor returns to the left edge of the viewport and down a row.
void vdc_text_move_cursor_right(void) {
	if (++vdc_text_cursor_position.x > vdc_text_viewport.right) {
		vdc_text_cursor_position.x = vdc_text_viewport.left;
		vdc_text_move_cursor_down();
	}
}

// Moves the cursor up a row. If it moves past the top edge of the viewport the cursor returns to the top edge and the viewport scrolls down one row.
void vdc_text_move_cursor_up(void) {
	if (--vdc_text_cursor_position.y < vdc_text_viewport.top) {
		vdc_text_cursor_position.y = vdc_text_viewport.top;
		vdc_text_scroll_down();
	}
}

// Moves the text cursor left a column. If it moves past the left edge of the viewport the cursor returns to the right edge of the viewport and up a row.
void vdc_text_move_cursor_left(void) {
	if (--vdc_text_cursor_position.x < vdc_text_viewport.left) {
		vdc_text_cursor_position.x = vdc_text_viewport.right;
		vdc_text_move_cursor_up();
	}
}

// Puts a character on the display at the current cursor position and advances the cursor.
void vdc_text_put_char(char c) {
	vdc_text_put_map(c);
	vdc_text_move_cursor_right();
}

// Set the text viewport. The viewport is clipped to the screen and if the cursor is no longer inside the viewport it is moved to the top-left of the new viewport.
void vdc_text_set_viewport(struct vdc_rectangle_t rect) {
	int x0, x1, y0, y1, clip;
	
	// Sort the coordinates.
	if (rect.left < rect.right) {
		x0 = rect.left;
		x1 = rect.right;
	} else {
		x0 = rect.right;
		x1 = rect.left;
	}
	if (rect.top < rect.bottom) {
		y0 = rect.top;
		y1 = rect.bottom;
	} else {
		y0 = rect.bottom;
		y1 = rect.top;
	}
	
	// Clip the points.
	if (x0 < 0) x0 = 0;
	if (x1 < 0) x1 = 0;
	if (y0 < 0) y0 = 0;
	if (y1 < 0) y1 = 0;
	if (y0 > 29) y0 = 29;
	if (y1 > 29) y1 = 29;
	clip = (320 / vdc_text_font_size) - 1;
	if (x0 > clip) x0 = clip;
	if (x1 > clip) x1 = clip;
	
	// Set the new viewport.
	vdc_text_viewport.left = x0;
	vdc_text_viewport.top = y0;
	vdc_text_viewport.right = x1;
	vdc_text_viewport.bottom = y1;
	
	// Clip the cursor.
	if (vdc_text_cursor_position.x < vdc_text_viewport.left || vdc_text_cursor_position.x > vdc_text_viewport.right || vdc_text_cursor_position.y < vdc_text_viewport.top || vdc_text_cursor_position.y > vdc_text_viewport.bottom) {
		vdc_text_cursor_position.x = vdc_text_viewport.left;
		vdc_text_cursor_position.y = vdc_text_viewport.top;
	}
}
